"""Definition of an ElkM1 Custom Value"""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, cp_encode, cw_encode


class Setting(Element):
    """Class representing an Custom Value"""
    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.value_format = None
        self.value = None

    def set(self, value):
        """(Helper) Set custom value."""
        self._elk.send(cw_encode(self._index, value, self.value_format))

class Settings(Elements):
    """Handling for multiple custom values"""
    def __init__(self, elk):
        super().__init__(elk, Setting, Max.SETTINGS.value)
        add_message_handler('CR', self._cr_handler)

    def sync(self):
        """Retrieve custom values from ElkM1"""
        self.elk.send(cp_encode())
        self.get_descriptions(TextDescriptions.SETTING.value)

    def _cr_handler(self, index, value, value_format):
        custom_value = self.elements[index]
        custom_value.value_format = value_format
        custom_value.value = value



"""Definition of an ElkM1 Task"""
from time import time

from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, tn_encode


class Task(Element):
    """Class representing an Task"""

    def __init__(self, index, elk): # pylint: disable=useless-super-delegation
        super().__init__(index, elk)
        self.last_change = None

    def activate(self):
        """(Helper) Activate task"""
        self._elk.send(tn_encode(self._index))

class Tasks(Elements):
    """Handling for multiple tasks"""
    def __init__(self, elk):
        super().__init__(elk, Task, Max.TASKS.value)
        add_message_handler('TC', self._tc_handler)

    def sync(self):
        """Retrieve tasks from ElkM1"""
        self.get_descriptions(TextDescriptions.TASK.value)

    def _tc_handler(self, task):
        self.elements[task].setattr('last_change', time())



"""Definition of an ElkM1 Thermostat"""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, tr_encode, ts_encode


class Thermostat(Element):
    """Class representing an Thermostat"""
    def __init__(self, index, elk): # pylint: disable=useless-super-delegation
        super().__init__(index, elk)
        self.mode = 0
        self.hold = False
        self.fan = 0
        self.current_temp = 0
        self.heat_setpoint = 0
        self.cool_setpoint = 0
        self.humidity = 0

    def set(self, element_to_set, value):
        """(Helper) Set thermostat"""
        self._elk.send(ts_encode(self._index, value, element_to_set))

class Thermostats(Elements):
    """Handling for multiple areas"""
    def __init__(self, elk):
        super().__init__(elk, Thermostat, Max.THERMOSTATS.value)
        add_message_handler('ST', self._st_handler)
        add_message_handler('TR', self._tr_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.get_descriptions(TextDescriptions.THERMOSTAT.value)

    def _got_desc(self, descriptions):
        super()._got_desc(descriptions)
        # Only poll thermostats that have a name defined
        for thermostat in self.elements:
            if not thermostat.is_default_name():
                self.elk.send(tr_encode(thermostat._index))

    def _st_handler(self, group, device, temperature):
        if group == 2:
            self.elements[device].setattr('current_temp', temperature)

    # pylint: disable=too-many-arguments
    def _tr_handler(self, thermostat_index, mode, hold, fan, current_temp,
                    heat_setpoint, cool_setpoint, humidity):
        thermostat = self.elements[thermostat_index]
        thermostat.setattr('mode', mode)
        thermostat.setattr('hold', hold)
        thermostat.setattr('fan', fan)
        thermostat.setattr('current_temp', current_temp)
        thermostat.setattr('heat_setpoint', heat_setpoint)
        thermostat.setattr('cool_setpoint', cool_setpoint)
        thermostat.setattr('humidity', humidity)



"""
Constants used across package
"""

from enum import Enum

class Max(Enum):
    """Max number of elements on the panel"""
    AREAS = 8
    COUNTERS = 64
    KEYPADS = 16
    OUTPUTS = 208
    SETTINGS = 20
    TASKS = 32
    THERMOSTATS = 16
    USERS = 203
    LIGHTS = 256
    ZONES = 208
    ZONE_TEMPS = 16

class ZoneType(Enum):
    """Types of Elk zones"""
    DISABLED = 0
    BURGLAR_ENTRY_EXIT_1 = 1
    BURGLAR_ENTRY_EXIT_2 = 2
    BURGLAR_PERIMETER_INSTANT = 3
    BURGLAR_INTERIOR = 4
    BURGLAR_INTERIOR_FOLLOWER = 5
    BURGLAR_INTERIOR_NIGHT = 6
    BURGLAR_INTERIOR_NIGHT_DELAY = 7
    BURGLAR24_HOUR = 8
    BURGLAR_BOX_TAMPER = 9
    FIRE_ALARM = 10
    FIRE_VERIFIED = 11
    FIRE_SUPERVISORY = 12
    AUX_ALARM_1 = 13
    AUX_ALARM_2 = 14
    KEYFOB = 15
    NON_ALARM = 16
    CARBON_MONOXIDE = 17
    EMERGENCY_ALARM = 18
    FREEZE_ALARM = 19
    GAS_ALARM = 20
    HEAT_ALARM = 21
    MEDICAL_ALARM = 22
    POLICE_ALARM = 23
    POLICE_NO_INDICATION = 24
    WATER_ALARM = 25
    KEY_MOMENTARY_ARM_DISARM = 26
    KEY_MOMENTARY_ARM_AWAY = 27
    KEY_MOMENTARY_ARM_STAY = 28
    KEY_MOMENTARY_DISARM = 29
    KEY_ON_OFF = 30
    MUTE_AUDIBLES = 31
    POWER_SUPERVISORY = 32
    TEMPERATURE = 33
    ANALOG_ZONE = 34
    PHONE_KEY = 35
    INTERCOM_KEY = 36

class ZonePhysicalStatus(Enum):
    """Zone physical status; name capitalized so can be used in displays"""
    UNCONFIGURED = 0
    OPEN = 1
    EOL = 2
    SHORT = 3

class ZoneLogicalStatus(Enum):
    """Zone logical status; name capitalized so can be used in displays"""
    NORMAL = 0
    TROUBLED = 1
    VIOLATED = 2
    BYPASSED = 3

class ArmedStatus(Enum):
    """Area arming status: armed status"""
    DISARMED = '0'
    ARMED_AWAY = '1'
    ARMED_STAY = '2'
    ARMED_STAY_INSTANT = '3'
    ARMED_TO_NIGHT = '4'
    ARMED_TO_NIGHT_INSTANT = '5'
    ARMED_TO_VACATION = '6'

class ArmUpState(Enum):
    """Area arming status: Ability to arm"""
    NOT_READY_TO_ARM = '0'
    READY_TO_ARM = '1'
    CAN_BE_FORCE_ARMED = '2'
    ARMED_AND_EXIT_TIMER_RUNNING = '3'
    FULLY_ARMED = '4'
    FORCE_ARMED = '5'
    ARMED_WITH_BYPASS = '6'

class AlarmState(Enum):
    """Area arming status: Current alarm state"""
    NO_ALARM_ACTIVE = '0'
    ENTRANCE_DELAY_ACTIVE = '1'
    ALARM_ABORT_DELAY_ACTIVE = '2'
    FIRE_ALARM = '3'
    MEDICAL_ALARM = '4'
    POLICE_ALARM = '5'
    BURGLAR_ALARM = '6'
    AUX_1_ALARM = '7'
    AUX_2_ALARM = '8'
    AUX_3_ALARM = '9'
    AUX_4_ALARM = ':'
    CARBON_MONOXIDE_ALARM = ';'
    EMERGENCY_ALARM = '<'
    FREEZE_ALARM = '='
    GAS_ALARM = '>'
    HEAT_ALARM = '?'
    WATER_ALARM = '@'
    FIRE_SUPERVISORY = 'A'
    VERIFY_FIRE = 'B'

class ArmLevel(Enum):
    """Levels for Arm/Disarm al_encode messages"""
    DISARM = '0'
    ARMED_AWAY = '1'
    ARMED_STAY = '2'
    ARMED_STAY_INSTANT = '3'
    ARMED_NIGHT = '4'
    ARMED_NIGHT_INSTANT = '5'
    ARMED_VACATION = '6'
    ARM_TO_NEXT_AWAY_MODE = '7'
    ARM_TO_NEXT_STAY_MODE = '8'
    FORCE_ARM_TO_AWAY_MODE = '9'
    FORCE_ARM_TO_STAY_MODE = ':'

class ZoneAlarmState(Enum):
    """Alarm state for a zone"""
    NO_ALARM = '0'
    BURGLAR_ENTRY_EXIT_1 = '1'
    BURGLAR_ENTRY_EXIT_2 = '2'
    BURGLAR_PERIMETER_INSTANT = '3'
    BURGLAR_INTERIOR = '4'
    BURGLAR_INTERIOR_FOLLOWER = '5'
    BURGLAR_INTERIOR_NIGHT = '6'
    BURGLAR_INTERIOR_NIGHT_DELAY = '7'
    BURGLAR_24_HOUR = '8'
    BURGLAR_BOX_TAMPER = '9'
    FIRE_ALARM = ':'
    FIRE_VERIFIED = ';'
    FIRE_SUPERVISORY = '<'
    AUX_ALARM_1 = '='
    AUX_ALARM_2 = '>'
    KEYFOB = '?'  # not used
    NON_ALARM = '@' # not used
    CARBON_MONOXIDE = 'A'
    EMERGENCY_ALARM = 'B'
    FREEZE_ALARM = 'C'
    GAS_ALARM = 'D'
    HEAT_ALARM = 'E'
    MEDICAL_ALARM = 'F'
    POLICE_ALARM = 'G'
    POLICE_NO_INDICATION = 'H'
    WATER_ALARM = 'I'

class TextDescriptions(Enum):
    """Types of description strings that can be retrieved from the panel"""
    ZONE = (0, Max.ZONES.value)
    AREA = (1, Max.AREAS.value)
    USER = (2, Max.USERS.value)
    KEYPAD = (3, Max.KEYPADS.value)
    OUTPUT = (4, 64)
    TASK = (5, Max.TASKS.value)
    TELEPHONE = 6
    LIGHT = (7, Max.LIGHTS.value)
    ALARM_DURATION = 8
    SETTING = (9, Max.SETTINGS.value)
    COUNTER = (10, Max.COUNTERS.value)
    THERMOSTAT = (11, Max.THERMOSTATS.value)
    FUNCTION_KEY_1 = 12
    FUNCTION_KEY_2 = 13
    FUNCTION_KEY_3 = 14
    FUNCTION_KEY_4 = 15
    FUNCTION_KEY_5 = 16
    FUNCTION_KEY_6 = 17
    AUDIO_ZONE = 18
    AUDIO_SOURCE = 19

# Map to convert message code to descriptive string
MESSAGE_MAP = {
    'AP': "Send ASCII String",
    'AR': "Alarm Reporting to Ethernet",
    'AS': "Arming status report data",
    'AT': "Ethernet Test to IP",
    'AZ': "Alarm by zone reply",
    'CA': "Reply Touchscreen audio command",
    'CC': "Control output change update",
    'CD': "Outgoing Audio Equip Command",
    'CR': "Custom value report data",
    'CS': "Control output status report data",
    'CU': "Change user code reply",
    'CV': "Counter Value Data",
    'DK': "Display KP LCD Data, not used",
    'DS': "Lighting Poll Response",
    'EE': "Entry/Exit Time Data",
    'EM': "Email Trigger to M1XEP",
    'IC': "Send invalid user code digits",
    'IE': "Installer program exited",
    'IP': "M1XSP Insteon Program",
    'IR': "M1XSP Insteon Read",
    'KA': "Keypad areas report data",
    'KC': "Keypad key change update",
    'KF': "Function key pressed data",
    'LD': "Log data with index",
    'LW': "Reply temperature data",
    'NS': "Reply Source Name",
    'NZ': "Reply Zone Name",
    'PC': "PLC change update",
    'PS': "PLC status report data",
    'RE': "Reset Ethernet Module",
    'RP': "ELKRP connected",
    'RR': "Real Time Clock Data",
    'SD': "Text string description report data",
    'SS': "System Trouble Status data",
    'ST': "Temperature report data",
    'T2': "Reply Omnistat 2 data",
    'TC': "Task change update",
    'TR': "Thermostat data report",
    'UA': "User code areas report data",
    'VN': "Reply Version Number of M1",
    'XB': "reserved by ELKRP",
    'XK': "Request Ethernet test",
    'ZB': "Zone bypass report data",
    'ZC': "Zone change update",
    'ZD': "Zone definition report data",
    'ZP': "Zone partition report data",
    'ZS': "Zone status report data",
    'ZV': "Zone analog voltage data",

    'a0': "Disarm",
    'a1': "Arm to away",
    'a2': "Arm to stay",
    'a3': "Arm to stay instant",
    'a4': "Arm to night",
    'a5': "Arm to night instant",
    'a6': "Arm to vacation",
    'a7': "Arm, step to next Away Mode",
    'a8': "Arm, step to next Stay Mode",
    'a9': "Force Arm to Away Mode",
    'a:': "Force Arm to Stay Mode",
    'ar': "Alarm Reporting Acknowledge",
    'as': "Request arming status",
    'at': "Ethernet Test Acknowledge",
    'az': "Alarm by zone request",
    'ca': "Request Touchscreen audio command",
    'cd': "Incoming Audio Equip Command",
    'cf': "Control output OFF",
    'cn': "Control output ON",
    'cp': "Request ALL custom values",
    'cr': "Request custom value",
    'cs': "Control output status request",
    'ct': "Control output TOGGLE",
    'cu': "Change user code request",
    'cv': "Request Counter value",
    'cw': "Write custom value data",
    'cx': "Write counter value",
    'dm': "Display message",
    'ds': "Lighting Poll Request",
    'ip': "M1XSP Insteon Program",
    'ir': "M1XSP Insteon Read",
    'ka': "Request keypad areas",
    'kc': "Request F Key illumination status",
    'kf': "Request simulated function key press",
    'ld': "Request log data, with index",
    'le': "Write Log Data Entry",
    'lw': "Request temperature data",
    'pc': "Control any PLC device",
    'pf': "Turn OFF PLC device",
    'pn': "Turn ON PLC device",
    'ps': "Request PLC status",
    'pt': "Toggle PLC device",
    'rr': "Request Real Time Clock Read",
    'rs': "Used by Touchscreen",
    'rw': "Real Time Clock Write",
    'sd': "Request text string descriptions",
    'sp': "Speak phrase",
    'ss': "Request System Trouble Status",
    'st': "Request temperature",
    'sw': "Speak word",
    't2': "Request Omnistat 2 data",
    'tn': "Task activation",
    'tr': "Request thermostat data",
    'ts': "Set thermostat data",
    'ua': "Request user code areas",
    'vn': "request Version Number of M1",
    'xk': "Reply from Ethernet test",
    'zb': "Zone bypass request",
    'zd': "Request zone definition data",
    'zp': "Zone partition request",
    'zs': "Zone status request",
    'zv': "Request Zone analog voltage",
}

class ThermostatSetting(Enum):
    """Thermostat consts when setting"""
    MODE = 0
    HOLD = 1
    FAN = 2
    GET_TEMPERATURE = 3
    COOL_SETPOINT = 4
    HEAT_SETPOINT = 5

class ThermostatMode(Enum):
    """Thermostat modes"""
    OFF = 0
    HEAT = 1
    COOL = 2
    AUTO = 3
    EMERGENCY_HEAT = 4

class ThermostatFan(Enum):
    """Thermostat fan"""
    AUTO = 0
    ON = 1

class ThermostatHold(Enum):
    """Thermostat hold"""
    OFF = 0
    ON = 1

class SettingFormat(Enum):
    NUMBER = 0
    TIMER = 1
    TIME_OF_DAY = 2

class ElkRPStatus(Enum):
    DISCONNECTED = 0
    CONNECTED = 1
    INITIALIZING = 2



"""Definition of an ElkM1 Area"""
import time

from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, as_encode, al_encode


class Area(Element):
    """Class representing an Area"""
    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.armed_status = None
        self.arm_up_state = None
        self.alarm_state = None
        self.is_exit = False
        self.timer1 = 0
        self.timer2 = 0
        self.timer_timerstamp = 0

    def arm(self, level, code):
        """(Helper) Arm system at specified level (away, vacation, etc)"""
        self._elk.send(al_encode(level, self._index, code))

    def disarm(self, code):
        """(Helper) Disarm system."""
        self.arm(0, code)

class Areas(Elements):
    """Handling for multiple areas"""
    def __init__(self, elk):
        super().__init__(elk, Area, Max.AREAS.value)
        add_message_handler('AS', self._as_handler)
        add_message_handler('EE', self._ee_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.elk.send(as_encode())
        self.get_descriptions(TextDescriptions.AREA.value)

    def _as_handler(self, armed_statuses, arm_up_states, alarm_states):
        for area in self.elements:
            area.setattr('armed_status', armed_statuses[area.index])
            area.setattr('arm_up_state', arm_up_states[area.index])
            area.setattr('alarm_state', alarm_states[area.index])

    def _ee_handler(self, area, is_exit, timer1, timer2, armed_status):
        area = self.elements[area]
        area.setattr('armed_status', armed_status)
        area.setattr('timer1', timer1)
        area.setattr('timer2', timer2)
        area.setattr('is_exit', is_exit)
        area.setattr('timer_timestamp', time.time())



"""Master class that combines all ElkM1 pieces together."""

import asyncio
import logging
from importlib import import_module
import serial_asyncio

from .message import message_decode
from .proto import Connection
from .util import call_sync_handlers, parse_url, url_scheme_is_secure


LOG = logging.getLogger(__name__)

class Elk:
    """Represents all the components on an Elk panel."""

    def __init__(self, config, loop=None):
        """Initialize a new Elk instance."""
        self.loop = loop if loop else asyncio.get_event_loop()
        self._config = config
        self._conn = None
        self.connection_lost_callbk = None
        self._connection_retry_timer = 1

        # Setup for all the types of elements tracked
        if 'element_list' in config:
            self.element_list = config['element_list']
        else:
            self.element_list = ['panel', 'zones', 'lights', 'areas',
                                 'tasks', 'keypads', 'outputs', 'thermostats',
                                 'counters', 'settings', 'users']
        for element in self.element_list:
            self._create_element(element)

    def _create_element(self, element):
        module = import_module('elkm1.'+element)
        class_ = getattr(module, element.capitalize())
        setattr(self, element, class_(self))

    async def _connect(self, connection_lost_callbk=None):
        """Asyncio connection to Elk."""
        self.connection_lost_callbk = connection_lost_callbk
        url = self._config['url']
        LOG.debug("Elk connect to %s", url)
        scheme, dest, param, ssl_context = parse_url(url)
        try:
            if scheme == 'serial':
                _coro = await serial_asyncio.create_serial_connection(self.loop,
                        lambda: Connection(self.loop,
                        self._connected, self._disconnected, self._got_data),
                        dest, baudrate=param)
            else:
                # pylint: disable=C0330
                _coro = await self.loop.create_connection(
                        lambda: Connection(self.loop,
                        self._connected, self._disconnected, self._got_data),
                        host=dest, port=param, ssl=ssl_context)
        except:
            LOG.debug("Connection failed. Retrying in {} seconds".format(
                self._connection_retry_timer))
            self.loop.call_later(self._connection_retry_timer, self.connect)
            self._connection_retry_timer = 2 * self._connection_retry_timer \
                if self._connection_retry_timer < 256 else 300

    def _connected(self, _transport, conn):
        """Login and sync the ElkM1 panel to memory."""
        self._conn = conn
        self._connection_retry_timer = 1
        if url_scheme_is_secure(self._config['url']):
            self._conn.write_data(self._config['userid'], raw=True)
            self._conn.write_data(self._config['password'], raw=True)
        call_sync_handlers()

    def _disconnected(self):
        # TODO: callbk out of library
        LOG.debug("elk disconnected callback")
        self._conn = None
        self.loop.call_later(self._connection_retry_timer, self.connect)

    def _got_data(self, data): # pylint: disable=no-self-use
        LOG.debug("got_data '%s'", data)
        try:
            message_decode(data)
        except ValueError as err:
            LOG.debug(err)

    def connect(self):
        """Connect to the panel"""
        coro = self._connect()
        asyncio.ensure_future(coro)

    def run(self):
        """Enter the asyncio loop."""
        self.loop.run_forever()

    def send(self, msg):
        """Send a message to Elk panel."""
        self._conn.write_data(msg.message, msg.response_command)

    def pause(self):
        self._conn.pause()

    def resume(self):
        self._conn.resume()



import pytest
from unittest.mock import Mock

from elkm1.message import add_message_handler, message_decode, MessageEncode, \
    housecode_to_index, index_to_housecode, ps_encode

def test_housecode_to_index_accepts_valid_codes():
    assert housecode_to_index('A01') == 0
    assert housecode_to_index('P16') == 255

def test_housecode_to_index_raises_error_on_invalid():
    with pytest.raises(ValueError): housecode_to_index('asdf')
    with pytest.raises(ValueError): housecode_to_index('a01')
    with pytest.raises(ValueError): housecode_to_index('Q01')
    with pytest.raises(ValueError): housecode_to_index('A00')
    with pytest.raises(ValueError): housecode_to_index('A17')

def test_index_to_housecode_accepts_valid_indices():
    assert index_to_housecode(0) == 'A01'
    assert index_to_housecode(16) == 'B01'
    assert index_to_housecode(255) == 'P16'

def test_index_to_housecode_raises_error_on_invalid():
    with pytest.raises(ValueError): index_to_housecode(-1)
    with pytest.raises(ValueError): index_to_housecode(256)

# def test_chsum_returns_correct_value():
#     assert cksum('09pfA0100') == 'BF'
#     assert cksum('08XXtest') == '28'
#     assert cksum('08test') == 'D8'

def test_decode_raises_value_error_on_bad_message():
    with pytest.raises(ValueError): message_decode('a really really bad message')

def test_decode_calls_unknown_handler_on_bad_command_or_not_implemented():
    mock_unknown_handler = Mock()
    add_message_handler('unknown', mock_unknown_handler)
    message_decode('08XXtest28')
    mock_unknown_handler.assert_called_once_with(msg_code='XX', data='test')

def test_decode_raises_value_error_on_length_too_long():
    with pytest.raises(ValueError) as excinfo:
        message_decode('42CV01000990030')
    assert str(excinfo.value) == 'Elk message length incorrect'

def test_decode_raises_value_error_on_length_too_short():
    with pytest.raises(ValueError) as excinfo:
        message_decode('02CV01000990030')
    assert str(excinfo.value) == 'Elk message length incorrect'

def test_decode_raises_value_error_on_bad_checksum():
    with pytest.raises(ValueError) as excinfo:
        message_decode('0DCV01000990042')
    assert str(excinfo.value) == 'Elk message checksum invalid'

def test_encode_message_with_a_variable():
    assert ps_encode(1) == ('07ps100', 'PS')




"""Definition of an ElkM1 Output"""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, cs_encode, cf_encode, \
                     cn_encode, ct_encode


class Output(Element):
    """Class representing an Output"""
    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.output_on = False

    def turn_off(self):
        """(Helper) Turn of an output"""
        self._elk.send(cf_encode(self._index))

    def turn_on(self, time):
        """(Helper) Turn on an output"""
        self._elk.send(cn_encode(self._index, time))

    def toggle(self):
        """(Helper) Toggle an output"""
        self._elk.send(ct_encode(self._index))

class Outputs(Elements):
    """Handling for multiple areas"""
    def __init__(self, elk):
        super().__init__(elk, Output, Max.OUTPUTS.value)
        add_message_handler('CC', self._cc_handler)
        add_message_handler('CS', self._cs_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.elk.send(cs_encode())
        self.get_descriptions(TextDescriptions.OUTPUT.value)

    def _cc_handler(self, output, output_status):
        self.elements[output].setattr('output_on', output_status)

    def _cs_handler(self, output_status):
        for output in self.elements:
            output.setattr('output_on', output_status[output.index])



"""
ElkM1 message encode/decode.

Message format:
    LLMM<data>00CC, where:
    LL - length in ASCII hex
    MM - message code
    <data> - message contents; length varies by message code
    00 - constant two characters reserved (almost all messages)
    CC - checksum

The panel numbers zones, keypads, etc. in base 1. The interface of
this module refers to them using base 0. Conversion to base 1 is done
on sending and conversion to base 0 is done on receiving. So, base 0/1
conversion is encapsulated in this module.
"""


from collections import namedtuple
import re

from .const import Max

MessageEncode = namedtuple('MessageEncode', ['message', 'response_command'])

_message_handlers = {} #pylint: disable=invalid-name

def add_message_handler(message_type, handler):
    """Manage callbacks for message handlers."""
    if message_type not in _message_handlers:
        _message_handlers[message_type] = []

    handlers = _message_handlers[message_type]
    if handler not in handlers:
        handlers.append(handler)

def housecode_to_index(housecode):
    """Convert a X10 housecode to a zero-based index"""
    match = re.search(r'^([A-P])(\d\d)$', housecode)
    if match:
        house_index = int(match.group(2))
        if house_index >= 1 and house_index <= 16:
            return (ord(match.group(1)) - ord('A')) * 16 + house_index - 1
    raise ValueError

def index_to_housecode(index):
    """Convert a zero-based index to a X10 housecode."""
    if index < 0 or index > 255:
        raise ValueError
    quotient, remainder = divmod(index, 16)
    return chr(quotient+ord('A')) + '{:02d}'.format(remainder+1)

def get_elk_command(line):
    """Return the 2 character command in the message."""
    if len(line) < 4:
        return ''
    return line[2:4]

def _status_decode(status):
    """Decode a 1 byte status into logical and physical statuses."""
    logical_status = (status & 0b00001100) >> 2
    physical_status = status & 0b00000011
    return (logical_status, physical_status)

class call_handlers(): # pylint: disable=invalid-name,too-few-public-methods
    """Decorator that calls out to message handlers"""
    def __init__(self, cmd):
        self.cmd = cmd

    def __call__(self, decode_func):
        def wrapped_f(msg):
            """Calls handlers"""
            decoded_msg = decode_func(msg)
            handlers = _message_handlers[self.cmd] if self.cmd in _message_handlers else []
            for handler in handlers:
                handler(**decoded_msg)
        return wrapped_f

@call_handlers('AS')
def _as_decode(msg):
    """AS: Arming status report."""
    return {'armed_statuses': [x for x in msg[4:12]],
            'arm_up_states': [x for x in msg[12:20]],
            'alarm_states': [x for x in msg[20:28]]}

@call_handlers('AZ')
def _az_decode(msg):
    """AZ: Alarm by zone report."""
    return {'alarm_status': [x for x in msg[4:4+Max.ZONES.value]]}

@call_handlers('CR')
def _cr_one_custom_value_decode(msg):
    index = int(msg[4:6])-1
    value = int(msg[6:11])
    value_format = int(msg[11])
    if value_format == 2:
        value = ((value>>8) & 0xff, value & 0xff)
    return {'index': index, 'value': value, 'value_format': value_format}

def _cr_decode(msg):
    """CR: Custom values"""
    if int(msg[4:6]) > 0:
        _cr_one_custom_value_decode(msg)
    else:
        part = 6
        for i in range(Max.SETTINGS.value):
            newmsg = '0ECR{cv:02d}{value:s}'.format(cv=i+1, value=msg[part:part+6])
            _cr_one_custom_value_decode(newmsg)
            part += 6

@call_handlers('CC')
def _cc_decode(msg):
    """CC: Output status for single output."""
    return {'output': int(msg[4:7])-1, 'output_status': msg[7] == '1'}

@call_handlers('CS')
def _cs_decode(msg):
    """CS: Output status for all outputs."""
    output_status = [x == '1' for x in msg[4:4+Max.OUTPUTS.value]]
    return {'output_status': output_status}

@call_handlers('CV')
def _cv_decode(msg):
    """CV: Counter value."""
    return {'counter': int(msg[4:6])-1, 'value': int(msg[6:11])}

@call_handlers('EE')
def _ee_decode(msg):
    """EE: Entry/exit timer report."""
    return {'area': int(msg[4:5])-1, 'is_exit': msg[5:6] == '0',
            'timer1': int(msg[6:9]), 'timer2': int(msg[9:12]),
            'armed_status': msg[12:13]}

@call_handlers('IC')
def _ic_decode(msg):
    """IC: Send Valid Or Invalid User Code Format."""
    code = 0
    start = 4
    for _ in range(6):
        code = code * 10 + int(msg[start:start+2])
        start += 2
    return {'code': code, 'user': int(msg[16:19])-1, 'keypad': int(msg[19:21])-1}

@call_handlers('IE')
def _ie_decode(_msg):
    """IE: Installer mode exited."""
    return {}

@call_handlers('KA')
def _ka_decode(msg):
    """KA: Keypad areas for all keypads."""
    return {'keypad_areas': [ord(x)-0x31 for x in msg[4:4+Max.KEYPADS.value]]}

@call_handlers('LW')
def _lw_decode(msg):
    """LW: temperatures from all keypads and zones 1-16."""
    keypad_temps = []
    zone_temps = []
    for i in range(16):
        keypad_temps.append(int(msg[4+3*i:7+3*i]) - 40)
        zone_temps.append(int(msg[52+3*i:55+3*i]) - 60)
    return {'keypad_temps': keypad_temps, 'zone_temps': zone_temps}

@call_handlers('PC')
def _pc_decode(msg):
    """PC: PLC (lighting) change."""
    housecode = msg[4:7]
    return {'housecode': housecode, 'index': housecode_to_index(housecode),
            'light_level': int(msg[7:9])}

@call_handlers('PS')
def _ps_decode(msg):
    """PS: PLC (lighting) status."""
    return {'bank': ord(msg[4]) - 0x30,
            'statuses': [ord(x)-0x30 for x in msg[5:69]]}

@call_handlers('RP')
def _rp_decode(msg):
    """RP: Remote programming status."""
    return {'remote_programming_status': int(msg[4:6])}

@call_handlers('SD')
def _sd_decode(msg):
    """SD: Description text."""
    desc_ch1 = msg[9]
    show_on_keypad = ord(desc_ch1) >= 0x80
    if show_on_keypad:
        desc_ch1 = chr(ord(desc_ch1) & 0x7f)
    return {'desc_type': int(msg[4:6]), 'unit': int(msg[6:9])-1,
            'desc': (desc_ch1+msg[10:25]).rstrip(),
            'show_on_keypad': show_on_keypad}

@call_handlers('ST')
def _st_decode(msg):
    """ST: Temperature update."""
    group = int(msg[4:5])
    temperature = int(msg[7:10])
    if group == 0:
        temperature -= 60
    elif group == 1:
        temperature -= 40

    return {'group': group, 'device': int(msg[5:7])-1, 'temperature': temperature}

@call_handlers('TC')
def _tc_decode(msg):
    """TC: Task change."""
    return {'task': int(msg[4:7])-1}

@call_handlers('TR')
def _tr_decode(msg):
    """TR: Thermostat data response."""
    return {'thermostat_index': int(msg[4:6])-1, 'mode': int(msg[6]),
            'hold': msg[7] == '1', 'fan': int(msg[8]),
            'current_temp': int(msg[9:11]), 'heat_setpoint': int(msg[11:13]),
            'cool_setpoint': int(msg[13:15]), 'humidity': int(msg[15:17])}

@call_handlers('VN')
def _vn_decode(msg):
    """VN: Version information."""
    elkm1_version = msg[4:10]
    xep_version = msg[10:16]
    return {'elkm1_version': elkm1_version, 'xep_version': xep_version}

@call_handlers('XK')
def _xk_decode(msg):
    """XK: Ethernet Test."""
    return {'real_time_clock': msg[4:20]}

@call_handlers('ZB')
def _zb_decode(msg):
    """ZB: Zone bypass report."""
    return {'zone_number': int(msg[4:7])-1, 'zone_bypassed': msg[7] == '1'}

@call_handlers('ZC')
def _zc_decode(msg):
    """ZC: Zone Change."""
    status = _status_decode(int(msg[7:8], 16))
    return {'zone_number': int(msg[4:7])-1, 'zone_status': status}

@call_handlers('ZD')
def _zd_decode(msg):
    """ZD: Zone definitions."""
    zone_definitions = [ord(x)-0x30 for x in msg[4:4+Max.ZONES.value]]
    return {'zone_definitions': zone_definitions}

@call_handlers('ZP')
def _zp_decode(msg):
    """ZP: Zone partitions."""
    zone_partitions = [ord(x)-0x31 for x in msg[4:4+Max.ZONES.value]]
    return {'zone_partitions': zone_partitions}

@call_handlers('ZS')
def _zs_decode(msg):
    """ZS: Zone statuses."""
    zone_statuses = [_status_decode(int(x, 16)) for x in msg[4:4+Max.ZONES.value]]
    return {'zone_statuses': zone_statuses}

@call_handlers('ZV')
def _zv_decode(msg):
    """ZV: Zone voltage."""
    return {'zone_number': int(msg[4:7])-1, 'zone_voltage': int(msg[7:10])/10}

@call_handlers('unknown')
def _unknown_decode(msg):
    """Generic handler called when no specific handler exists"""
    return {'msg_code': msg[2:4], 'data': msg[4:-2]}

@call_handlers('timeout')
def timeout_decode(msg_code):
    """Called directly when a timeout happens when response not received"""
    return {'msg_code': msg_code}

def _check_checksum(msg):
    """Ensure checksum in message is good."""
    checksum = int(msg[-2:], 16)
    for char in msg[:-2]:
        checksum += ord(char)
    if (checksum % 256) != 0:
        raise ValueError("Elk message checksum invalid")

def _check_message_valid(msg):
    """Check packet length valid and that checksum is good."""
    try:
        if int(msg[:2], 16) != (len(msg) - 2):
            raise ValueError("Elk message length incorrect")
        _check_checksum(msg)
    except IndexError:
        raise ValueError("Elk message length incorrect")

def message_decode(msg):
    """Decode an Elk message by passing to appropriate decoder"""
    _check_message_valid(msg)

    cmd = msg[2:4]
    if not cmd.isupper():
        raise ValueError("Invalid message code")

    decoder_name = '_' + cmd.lower() + '_decode'
    try:
        if not callable(globals()[decoder_name]):
            raise ValueError
    except (KeyError, ValueError):
        _unknown_decode(msg)
        return
    globals()[decoder_name](msg)

def al_encode(arm_mode, area, user_code):
    """al: Arm system. Note in 'al' the 'l' can vary"""
    # TODO: AS message not always returned; IC msg could be returned
    return MessageEncode('0Da{}{:1d}{:06d}00'.format(arm_mode, area+1,
                                                     user_code), 'AS')

def as_encode():
    """as: Get area status."""
    return MessageEncode('06as00', 'AS')

def az_encode():
    """az: Get alarm by zone."""
    return MessageEncode('06az00', 'AZ')

def cf_encode(output):
    """cf: Turn off output."""
    return MessageEncode('09cf{:03d}00'.format(output+1), None)

def ct_encode(output):
    """ct: Toggle output."""
    return MessageEncode('09ct{:03d}00'.format(output+1), None)

def cn_encode(output, time):
    """cn: Turn on output."""
    return MessageEncode('0Ecn{:03d}{:05d}00'.format(output+1, time), None)

def cs_encode():
    """cs: Get all output status."""
    return MessageEncode('06cs00', 'CS')

def cp_encode():
    """cp: Get ALL custom values."""
    return MessageEncode('06cp00', 'CR')

def cr_encode(index):
    """cr: Get a custom value."""
    return MessageEncode('08cr{cv:02d}00'.format(cv=index+1), 'CR')

def cw_encode(index, value, value_format):
    """cw: Write a custom value."""
    if value_format == 2:
        value = value[0] << 8 + value[1]
    return MessageEncode('0Dcw{:02d}{:05d}00'.format(index+1, value), None)

def cv_encode(counter):
    """cv: Get counter."""
    return MessageEncode('08cv{c:02d}00'.format(c=counter+1), 'CV')

def cx_encode(counter, value):
    """cx: Change counter value."""
    return MessageEncode('0Dcx{:02d}{:05d}00'.format(counter+1, value), 'CV')

# pylint: disable=too-many-arguments
def dm_encode(keypad_area, clear, beep, timeout, line1, line2):
    """dm: Display message on keypad."""
    return MessageEncode('2Edm{:1d}{:1d}{:1d}{:05d}{:^<16.16}{:^<16.16}00'.format(
        keypad_area+1, clear, beep, timeout, line1, line2), None)

def ka_encode():
    """ka: Get keypad areas."""
    return MessageEncode('06ka00', 'KA')

def lw_encode():
    """lw: Get temperature data."""
    return MessageEncode('06lw00', 'LW')

def pc_encode(index, function_code, extended_code, time):
    """pc: Control any PLC device."""
    return MessageEncode('11pc{hc}{fc:02d}{ec:02d}{time:04d}00'.
                         format(hc=index_to_housecode(index),
                                fc=function_code, ec=extended_code,
                                time=time), None)

def pf_encode(index):
    """pf: Turn off light."""
    return MessageEncode('09pf{}00'.format(index_to_housecode(index)), None)

def pn_encode(index):
    """pn: Turn on light."""
    return MessageEncode('09pn{}00'.format(index_to_housecode(index)), None)

def ps_encode(bank):
    """ps: Get lighting status."""
    return MessageEncode('07ps{:1d}00'.format(bank), 'PS')

def pt_encode(index):
    """pt: Toggle light."""
    return MessageEncode('09pt{hc}00'.format(hc=index_to_housecode(index)), None)

def sd_encode(desc_type, unit):
    """sd: Get description."""
    return MessageEncode('0Bsd{:02d}{:03d}00'.format(desc_type, unit+1), 'SD')

def sp_encode(phrase):
    """sp: Speak phrase."""
    return MessageEncode('09sp{:03d}00'.format(phrase), None)

def sw_encode(word):
    """sp: Speak word."""
    return MessageEncode('09sp{:03d}00'.format(word), None)

def tn_encode(task):
    """tn: Activate task."""
    return MessageEncode('09tn{:03d}00'.format(task+1), None)

def tr_encode(thermostat):
    """tr: Request thermostat data."""
    return MessageEncode('08tr{:02d}00'.format(thermostat+1), None)

def ts_encode(thermostat, value, element):
    """ts: Set thermostat data."""
    return MessageEncode('0Bts{:02d}{:02d}{:1d}00'.format(
        thermostat+1, value, element), None)

def vn_encode():
    """zd: Get panel software version information."""
    return MessageEncode('06vn00', 'VN')

def zb_encode(zone, area, user_code):
    """zb: Zone bypass request. Zone < 0 unbypass all; Zone > Max bypass all."""
    if zone < 0:
        zone = 0
    elif zone > Max.ZONES.value:
        zone = 999
    else:
        zone += 1
    return MessageEncode('10zb{zone:03d}{area:1d}{code:06d}00'.format(
        zone=zone, area=area+1, code=user_code), 'ZB')

def zd_encode():
    """zd: Get zone definitions"""
    return MessageEncode('06zd00', 'ZD')

def zp_encode():
    """zp: Get zone partitions"""
    return MessageEncode('06zp00', 'ZP')

def zs_encode():
    """zs: Get zone statuses"""
    return MessageEncode('06zs00', 'ZS')

def zv_encode(zone):
    """zv: Get zone voltage"""
    return MessageEncode('09zv{zone:03d}00'.format(zone=zone+1), 'ZV')



"""Definition of an ElkM1 Area"""
from .const import ElkRPStatus
from .elements import Element
from .message import add_message_handler, vn_encode, lw_encode
from .util import add_sync_handler, call_sync_handlers


class Panel(Element):
    """Class representing an Area"""
    def __init__(self, elk):
        super().__init__(0, elk)
        self.real_time_clock = None
        self.elkm1_version = None
        self.xep_version = None
        self.remote_programming_status = 0
        self.setattr('name', 'ElkM1')
        add_sync_handler(self.sync)

    def sync(self):
        """Retrieve panel information from ElkM1"""
        add_message_handler('VN', self._vn_handler)
        add_message_handler('XK', self._xk_handler)
        add_message_handler('RP', self._rp_handler)
        add_message_handler('IE', call_sync_handlers)
        self._elk.send(vn_encode())
        self._elk.send(lw_encode())

    def speak_word(self, word):
        """(Helper) Speak word"""
        self._elk.send(sw_encode(self._index, word))

    def speak_phrase(self, phrase):
        """(Helper) Speak phrase"""
        self._elk.send(sp_encode(self._index, phrase))

    def _vn_handler(self, elkm1_version, xep_version):
        self.setattr('elkm1_version', elkm1_version)
        self.setattr('xep_version', xep_version)

    def _xk_handler(self, real_time_clock):
        self.setattr('real_time_clock', real_time_clock)

    def _rp_handler(self, remote_programming_status):
        if remote_programming_status == ElkRPStatus.DISCONNECTED.value:
            self._elk.resume()
        else:
            self._elk.pause()
        self.setattr('remote_programming_status', remote_programming_status)



"""Definition of an ElkM1 Custom Value"""
from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import add_message_handler, cx_encode, cv_encode


class Counter(Element):
    """Class representing an Counter"""
    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.value = None

    def set(self, value):
        """(Helper) Set counter to value"""
        self._elk.send(cx_encode(self._index, value))

# pylint: disable=R0903
class Counters(Elements):
    """Handling for multiple counters"""
    def __init__(self, elk):
        super().__init__(elk, Counter, Max.COUNTERS.value)
        add_message_handler('CV', self._cv_handler)

    def sync(self):
        """Retrieve values from ElkM1 on demand"""
        self.get_descriptions(TextDescriptions.COUNTER.value)

    def _got_desc(self, descriptions):
        super()._got_desc(descriptions)
        # Only poll counters that have a name defined
        for counter in self.elements:
            if not counter.is_default_name():
                self.elk.send(cv_encode(counter._index))

    def _cv_handler(self, counter, value):
        countr = self.elements[counter]
        countr.value = value
